# ----------------------------------------------------------------------------
# GLOBAL DATA:
set dbHandle ""
set dbResultHandle ""


# ------------------------------------------------------------------------------
define_compound_tag "<sql-query>compound</sql-query>"
proc process_SQL-QUERY args {
    global global_options dbcErrorMessage  dbHandle dbResultHandle

    array set params {
        COMPOUND        ""
        STMT    ""
        RESSECTION  "RESSECTION"
        RESENTRY        "RESENTRY"
        RESEMPTY        "RESEMPTY"
        RESERROR "RESERROR"
    }

	array set params $args

        set entry $params(RESENTRY)
        set section $params(RESSECTION)
        set compound [string trim $params(COMPOUND)]
        set stmt [formatStmt $params(STMT)]
        set res ""


        if { ![dbopen] } {
                append res "<$params(RESERROR)
                USERMSG=\"Database not open\" SYSTEMMSG=\"$dbcErrorMessage\">"
                return "[parse_html_text $res]"
        }


# aufsplitten des statements in seine einzelnen bestandteile
        set stmtElems [split $stmt]

        set fromIndex [lsearch $stmtElems from]
        set whereIndex [getWhereIndex $stmtElems]
        set subqueryIndex [getSubqueryIndex $stmtElems]

        set select [join [lrange $stmtElems 1 [expr $fromIndex-1]]]
        set from  [join [lrange $stmtElems [expr $fromIndex+1] [expr $whereIndex-1]]]


# bestimmen der einzelnen spaltennamen inkl. aufloesen des * bei einer select anfrage
        set selectElems {}
        set selectElemsTemp [split [string trim $select] ,]

        foreach e $selectElemsTemp {
           if {$e == "*"} {
                # die zu verarbeitenden tabellennamen
                set fromElemsTemp [split [string trim $from] ,]
                foreach fromE $fromElemsTemp {
                        dbquery "describe $fromE;"

                        while {[set row [sql fetchrow $dbHandle]] != ""} {
                                lappend selectElems [lindex $row 0]
                        }
                        sql endquery $dbHandle $dbResultHandle
                }
           } else {
                lappend selectElems $e
           }
        }


# existiert ein subquery index?
if {$subqueryIndex != -1} {

# es wurde eine subquery eingeleitet, es ist aber kein <sql-subquery>-tag vorhanden
        if {![regexp ".*<sql-subquery.*>.*" $compound ]} {
                append res "<$params(RESERROR)
                USERMSG=\"subquery specified in sql-statement, but no subquery-tag was found\" SYSTEMMSG=\"\">"
                return "[parse_html_text $res]"
        }

# wird diese zuerst einmal verarbeitet:
# ressub ist dabei die ergebnismenge der sub-query als liste
        set resSub [parse_xml_text $compound]

# eine leere subquery menge fuehrt zu einem leeren result-set
        if {$resSub == ""} {
                append res "<$params(RESEMPTY)>"
                return "[parse_html_text $res]"
        }


# der where teil wird aufgesplittet um das ergebnis der subquery mit einzubauen
        set oldWhere [join [lrange $stmtElems $whereIndex [expr $subqueryIndex-1]]]
        set where [createNewWhereCond $oldWhere $resSub]

# der rest des statements, z.b. "order by ... limit ...;"
        set rest [join [lrange $stmtElems [expr $subqueryIndex+1] [llength $stmtElems]]]
} else {
# keine subquery vorhanden
# die where bedingung muss nicht weiter betrachtet werden, es kann alles zum rest zusammengefasst werden
        set where ""
        set rest [join [lrange $stmtElems $whereIndex [llength $stmtElems]]]
}

# das gesamte neue statement zusammensetzen
        set stmt "select $select from $from $where $rest"


# ausfuehren der query
        set numRows [dbquery $stmt]
        if { $numRows < 0 } {
                append res "<$params(RESERROR)
		USERMSG=\"error in sql syntax: $stmt\" SYSTEMMSG=\"$dbcErrorMessage\">"
                return "[parse_html_text $res]"
        } elseif {$numRows == 0} {
                append res "<$params(RESEMPTY)>"
                return "[parse_html_text $res]"
        }

#start tag RESSECTION
        append res "<$section EFFROWS=\"$numRows\">"


    while {[set row [sql fetchrow $dbHandle]] != ""} {
        set resparams ""
        set n 0
# durchlaufen aller spalten einer ergebniszeile und zusammensetzen des ergebnisses mit den entsprechenden spaltennamen
        foreach e $row {
           append resparams "[lindex $selectElems $n]=$e "
           set n [expr ($n+1)]
        }

        append res "<$entry $resparams>"
    }

    sql endquery $dbHandle $dbResultHandle

#end tag RESSECTION
        append res "</$section>"



   if {![dbclose] } {
        append res "<$params(RESERROR)
		USERMSG=\"Unable to close Database\" SYSTEMMSG=\"$dbcErrorMessage\">"
        return "[parse_html_text $res]"
    }

	return [parse_html_text $res]
}



# ------------------------------------------------------------------------------
define_compound_tag "<sql-subquery>compound</sql-subquery>"
proc process_SQL-SUBQUERY args {
    global global_options dbcErrorMessage  dbHandle dbResultHandle

	trc process_SQL-SUBQUERY "starting"

        array set params {
                COMPOUND        ""
                STMT    ""
        }

	array set params $args

        set compound [string trim $params(COMPOUND)]
        set stmt [formatStmt $params(STMT)]

# aufsplitten des statements in seine einzelnen bestandteile
                set stmtElems [split $stmt]
                set whereIndex [getWhereIndex $stmtElems]
                set subqueryIndex [getSubqueryIndex $stmtElems]

# existiert eine subquery?
        if {$subqueryIndex != -1} {

# es wurde eine subquery eingeleitet, es ist aber kein <sql-subquery>-tag vorhanden
                if {![regexp ".*<sql-subquery.*>.*" $compound ]} {
                        append res "<$params(RESERROR)
                        USERMSG=\"subquery specified in sql-statement, but no subquery-tag was found\" SYSTEMMSG=\"\">"
                        return "[parse_html_text $res]"
                }

# wird diese zuerst einmal verarbeitet:
# ressub ist dabei die ergebnismenge der sub-query als liste
                set resSub [parse_xml_text $params(COMPOUND)]

                if {$resSub == {}} {
                        return $resSub
                }

# der teil bis zu "where"
                set first [join [lrange $stmtElems 0 [expr $whereIndex-1]]]

# der where teil wird aufgesplittet um das ergebnis der subquery mit einzubauen
                set oldWhere [join [lrange $stmtElems $whereIndex [expr $subqueryIndex-1]]]
                set where [createNewWhereCond $oldWhere $resSub]

# der rest des statements, z.b. "order by ... limit ...;"
                set rest [join [lrange $stmtElems [expr $subqueryIndex+1] [llength $stmtElems]]]

# das gesamte neue statement zusammensetzen
                set stmt "$first $where $rest"
        }

    set numRows [dbquery $stmt]
    set result [dbfetch]
    sql endquery $dbHandle $dbResultHandle

return $result
}


# ------------------------------------------------------------------------------
proc process_SQL-WRITE args {
    global global_options dbcErrorMessage  dbHandle dbResultHandle

       array set params {
                RESWRITE  "RESWRITE"
                RESERROR "RESERROR"
                STMT    ""
        }

	array set params $args

        set stmt [formatStmt $params(STMT)]
        set res ""

        if { ![dbopen] } {
                append res "<$params(RESERROR)
                USERMSG=\"Database not open\" SYSTEMMSG=\"$dbcErrorMessage\">"
                return "[parse_html_text $res]"
        }

        if { [catch {set dbResultHandle [sql exec $dbHandle $stmt]} dbcErrorMessage] } {
                append res "<$params(RESERROR)
		USERMSG=\"error in sql syntax: $stmt\" SYSTEMMSG=\"$dbcErrorMessage\">"
        } else {
                append res "<$params(RESWRITE)>"
        }

        sql endquery $dbHandle $dbResultHandle

        if {![dbclose] } {
                append res "<$params(RESERROR)
		USERMSG=\"Unable to close Database\" SYSTEMMSG=\"$dbcErrorMessage\">"
        }

        return "[parse_html_text $res]"
}




# ----------------------------------------------------------------------------
# PROC:    getWhereIndex
# PARAMS:  stmtElems: liste der einzelnen elemente eines sql-statements
# RETURN:  index position des ersten bedingungselementes
# ----------------------------------------------------------------------------
proc getWhereIndex {stmtElems} {
        return [lsearch -regexp $stmtElems {(where|group|order|limit|;)}]
}


# ----------------------------------------------------------------------------
# PROC:    getSubqueryIndex
# PARAMS:  stmtElems: liste der einzelnen elemente eines sql-statements
# RETURN:  index position des <subquery> elementes
# ----------------------------------------------------------------------------
proc getSubqueryIndex {stmtElems} {
        return [lsearch $stmtElems <subquery>]
}

# ----------------------------------------------------------------------------
# PROC:    formatStmt
#          entfernen aller unnuetzen (mehrfachen) leerzeichen aus dem sql-statement sowie umwandlung in kleinbuchstaben
#          notwendig fuer die weiteren vergleiche, da immer auf kleinbuchstaben verglichen wird
#          das abschliessende ";" wird durch ein leerzeichen von statement getrennt,
#          notwendig fuer die einfachere verarbeitung des statements
# PARAMS:  stmt: sql-statement
# RETURN:  formatiertes sql-statement
# ----------------------------------------------------------------------------
proc formatStmt {stmt} {
        regsub -all {  *} [string trim [string tolower $stmt]] " " stmt
        regsub ";" $stmt " ;" stmt
        return $stmt
}


# ----------------------------------------------------------------------------
# PROC:    createNewWhereCond
#          erzeugen einer neuen where bedingung durch zusammensetzen der ergebnismenge der unterabfrage
#          mit boolschen ausdruecken
# PARAMS:  where: die zu verarbeitende where bedingung
#          resSub: ergebnismenge der unterabfrage
# RETURN:  neue where bedingung
# ----------------------------------------------------------------------------
proc createNewWhereCond {where resSub} {
# aufspliten der where bedingung nach leerzeichen um die einzelnen elemente verarbeiten zu koennen
        set whereCondElems [split [string trim $where]]

# mindestens das letzte element ist die bedingung, z.b. in, all, any, <=, =, usw.
        set cond [lindex $whereCondElems [expr [llength $whereCondElems] -1]]


# bestimmen der neuen bedingungen fuer die alte unterabfrage. dabei sei:
# w1,.., wn elemente der unterabfrage (u)
# k eine spalte der abfrage
# dann gilt fuer (siehe fallunterscheidung)
switch -regexp $cond {
        in {    # der vergleich auf "not in" kann somit erst im zweiten schritt durchgefuehrt werden
                if {[lindex $whereCondElems [expr [llength $whereCondElems] -2]] == "not"} {
                        # k not in (u)   <=>   (k!=w1) and (k!=w2) and ...
                        set column [lindex $whereCondElems [expr [llength $whereCondElems] -3]]
                        set op "!="
                        set bool "and"
                        # bis zu welchen punkt kann der alte "where"-teil uebernommen werden
                        # zb: where id = 32 and  |  elem not in     => -4 (von hinten gezaehlt)
                        set endpoint "-4"
                } else {
                        # k in (u)   <=>   (k==w1) or (k==w2) or ...
                        set column [lindex $whereCondElems [expr [llength $whereCondElems] -2]]
                        set op "="
                        set bool "or"
                        set endpoint "-3"
                }
        }
        (=|<|>|<=|>=) {
                # k cond (u)   <=>   (k cond w1) and (k cond w2) and ...
                set column [lindex $whereCondElems [expr [llength $whereCondElems] -2]]
                set op $cond
                set bool "and"
                set endpoint "-3"
        }
        (all|any) {
                # k = any (u)   <=>   k in (u)
                # k = all (u)   <=>   k not in (u)
                # alle anderen vergleiche sind entsprechend
                set column [lindex $whereCondElems [expr [llength $whereCondElems] -3]]
                set op [lindex $whereCondElems [expr [llength $whereCondElems] -2]]
                set endpoint "-4"
                if {$cond == "all"} {set bool "and"
                } else {set bool "or"}
        }
        default {
                append res "<$params(RESERROR)
                USERMSG=\"Not supported sub-query condition: $cond\" SYSTEMMSG=\"\">"
                return "[parse_html_text $res]"
        }
}

#zusammensetzen der neuen bedingung
        set newCond "("
#durchlaufen der ergebnismenge der unterabfrage und zusammensetzen mit den oben bestimmten bedingungen
        foreach e $resSub {
                append newCond "($column$op\"$e\") $bool "
        }
# das letzte "bool" ist zuviel und muss wieder weg
        set newCond "[string range $newCond 0 [expr [string last $bool $newCond]-2]])"

# den unveraenderten alte "where"-teil noch davor
        set newWhere "[lrange $whereCondElems 0 [expr [llength $whereCondElems] $endpoint]] $newCond"

        return $newWhere
}




# ----------------------------------------------------------------------------
# PROC:    dbcopen
# PARAMS:  --
# RETURN:  1    - ok
#          else - error (see dbcErrorMessage)
# ----------------------------------------------------------------------------
proc dbopen {} {
    global dbHandle dbcErrorMessage global_options

    set dbName $global_options(DBNAME)
    set dbUser $global_options(DBUSER)
    set dbHost $global_options(DBHOST)
    set dbPass $global_options(DBPASS)

    # get DB handle
    if { [catch {set dbHandle [sql connect $dbHost $dbUser $dbPass]} dbcErrorMessage] } {
         return 0
    }

    if { [catch {sql selectdb $dbHandle $dbName} dbcErrorMessage] } {
	return 0
    }

    return 1
}


# ----------------------------------------------------------------------------
# PROC:    dbquery
# PARAMS:  stmt:  sql-statement
# RETURN:  >= 0: anzahl der selektierten zeilen
#          -1: error
# ----------------------------------------------------------------------------
proc dbquery {stmt} {

    global dbResultHandle dbHandle dbcErrorMessage

    if { [catch {set dbResultHandle [sql query $dbHandle $stmt]} dbcErrorMessage] } {
	return -1
    }
    if { [catch {set numRows [sql numrows $dbHandle $dbResultHandle]} dbcErrorMessage] } {
	return -1
    }
    return $numRows
}


# ----------------------------------------------------------------------------
# PROC:    dbfetch
# PARAMS:  --
# RETURN:  liste der selektierten zeilen
# ----------------------------------------------------------------------------
proc dbfetch {} {
    global dbHandle dbcErrorMessage

    set resultList ""
    while {[set row [sql fetchrow $dbHandle]] != ""} {
            lappend resultList $row
    }

    return $resultList
}


# ----------------------------------------------------------------------------
# PROC:    dbcclose
# PARAMS:  --
# RETURN:  1   - ok
#          else - error (see dbcErrorMessage)
# ----------------------------------------------------------------------------
proc dbclose {} {
    global dbHandle dbcErrorMessage

    if { [catch {sql disconnect $dbHandle} dbcErrorMessage] } {
        return 0
    }
    return 1
}